""" Utilities - Helper functions """

import collections
import os
import pickle
import re
import urllib.parse

import bs4
import dateutil.parser
import measurement.measures
import numpy as np
import pandas as pd
import pkg_resources
import rapidjson
import requests

# ====================================================================================================================
""" Change directory """


# Change directory to "dat" and sub-directories
def cd_dat(*sub_dir, dat_dir="dat", mkdir=False):
    path = pkg_resources.resource_filename(__name__, dat_dir)
    for x in sub_dir:
        path = os.path.join(path, x)
    if mkdir:
        os.makedirs(path, exist_ok=True)
    return path


# ====================================================================================================================
""" Save data """


# Save Pickle file
def save_pickle(pickle_data, path_to_pickle, verbose=True):
    """
    :param pickle_data: any object that could be dumped by the 'pickle' package
    :param path_to_pickle: [str] local file path
    :param verbose: [bool]
    :return: whether the data has been successfully saved
    """
    pickle_filename = os.path.basename(path_to_pickle)
    pickle_dir = os.path.basename(os.path.dirname(path_to_pickle))
    pickle_dir_parent = os.path.basename(os.path.dirname(os.path.dirname(path_to_pickle)))

    if verbose:
        print("{} \"{}\" ... ".format("Updating" if os.path.isfile(path_to_pickle) else "Saving",
                                      " - ".join([pickle_dir_parent, pickle_dir, pickle_filename])), end="")

    try:
        os.makedirs(os.path.dirname(os.path.abspath(path_to_pickle)), exist_ok=True)
        pickle_out = open(path_to_pickle, 'wb')
        pickle.dump(pickle_data, pickle_out)
        pickle_out.close()
        print("Successfully.") if verbose else None
    except Exception as e:
        print("Failed. {}.".format(e))


# Save JSON file
def save_json(json_data, path_to_json, verbose=True):
    """
    :param json_data: any object that could be dumped by the 'json' package
    :param path_to_json: [str] local file path
    :param verbose: [bool]
    :return: whether the data has been successfully saved
    """
    json_filename = os.path.basename(path_to_json)
    json_dir = os.path.basename(os.path.dirname(path_to_json))
    json_dir_parent = os.path.basename(os.path.dirname(os.path.dirname(path_to_json)))

    print("{} \"{}\" ... ".format("Updating" if os.path.isfile(path_to_json) else "Saving",
                                  " - ".join([json_dir_parent, json_dir, json_filename])), end="") if verbose else None
    try:
        os.makedirs(os.path.dirname(os.path.abspath(path_to_json)), exist_ok=True)
        json_out = open(path_to_json, 'w')
        rapidjson.dump(json_data, json_out)
        json_out.close()
        print("Successfully.") if verbose else None
    except Exception as e:
        print("Failed. {}.".format(e))


# ====================================================================================================================
""" Converter """


# Convert "miles.chains" to Network Rail mileages
def mile_chain_to_nr_mileage(miles_chains):
    """
    :param miles_chains: [str] 'miles.chains'
    :return: [str] 'miles.yards'

    Note on the 'ELRs and mileages' web page that 'mileages' are given in the form 'miles.chains'.
    """

    if pd.notna(miles_chains) and miles_chains != '':
        miles, chains = str(miles_chains).split('.')
        yards = measurement.measures.Distance(chain=chains).yd
        network_rail_mileage = '%.4f' % (int(miles) + round(yards / (10 ** 4), 4))
    else:
        network_rail_mileage = miles_chains
    return network_rail_mileage


# Convert Network Rail mileages to "miles.chains"
def nr_mileage_to_mile_chain(str_mileage):
    """
    :param str_mileage: [str] 'miles.yards'
    :return: [str] 'miles.chains'

    Note on the 'ELRs and mileages' web page that 'mileages' are given in the form 'miles.chains'.
    """

    if pd.notna(str_mileage) and str_mileage != '':
        miles, yards = str(str_mileage).split('.')
        chains = measurement.measures.Distance(yard=yards).chain
        miles_chains = '%.2f' % (int(miles) + round(chains / (10 ** 2), 2))
    else:
        miles_chains = str_mileage
    return miles_chains


# Convert str type mileage to numerical type
def str_to_num_mileage(str_mileage):
    return '' if str_mileage == '' else round(float(str_mileage), 4)


# Convert mileage to str type
def nr_mileage_num_to_str(x):
    return '%.4f' % round(float(x), 4)


# Convert yards to Network Rail mileages
def yards_to_nr_mileage(yards):
    if not pd.isnull(yards) and yards is not None:
        mileage_mi = np.floor(measurement.measures.Distance(yd=yards).mi)
        mileage_yd = int(yards - measurement.measures.Distance(mi=mileage_mi).yd)
        # Example: "%.2f" % round(2606.89579999999, 2)
        mileage = str('%.4f' % round((mileage_mi + mileage_yd / (10 ** 4)), 4))
        return mileage


# Convert Network Rail mileages to yards
def nr_mileage_to_yards(nr_mileage):
    if isinstance(nr_mileage, float):
        nr_mileage = str('%.4f' % nr_mileage)
    elif isinstance(nr_mileage, str):
        pass
    miles = int(nr_mileage.split('.')[0])
    yards = int(nr_mileage.split('.')[1])
    yards += measurement.measures.Distance(mi=miles).yd
    return yards


# Convert calendar year to Network Rail financial year
def year_to_financial_year(date):
    financial_date = date + pd.DateOffset(months=-3)
    return financial_date.year


# ====================================================================================================================
""" Parser """


# Get a list of parsed HTML tr's
def parse_tr(header, trs):
    """
    :param header: [list] list of column names of a requested table
    :param trs: [list] contents under tr tags of the web page
    :return: [list] list of lists each comprising a row of the requested table

    Get a list of parsed contents of tr-tag's, each of which corresponds to a piece of record
    *This is a key function to drive its following functions
    Reference: stackoverflow.com/questions/28763891/what-should-i-do-when-tr-has-rowspan

    """
    tbl_lst = []
    for row in trs:
        data = []
        for dat in row.find_all('td'):
            txt = dat.get_text()
            if '\n' in txt:
                t = txt.split('\n')
                txt = '%s (%s)' % (t[0], t[1:]) if '(' not in txt and ')' not in txt else '%s %s' % (t[0], t[1:])
                data.append(txt)
            else:
                data.append(txt)
        tbl_lst.append(data)

    row_spanned = []
    for no, tr in enumerate(trs):
        for td_no, rho in enumerate(tr.find_all('td')):
            # print(data.has_attr("rowspan"))
            if rho.has_attr('rowspan'):
                row_spanned.append((no, int(rho['rowspan']), td_no, rho.text))

    if row_spanned:
        d = collections.defaultdict(list)
        for k, *v in row_spanned:
            d[k].append(v)
        row_spanned = list(d.items())

        for x in row_spanned:
            i, to_repeat = x[0], x[1]
            for y in to_repeat:
                for j in range(1, y[0]):
                    if y[2] in tbl_lst[i] and y[2] != '\xa0':
                        y[1] += pd.np.abs(tbl_lst[i].index(y[2]) - y[1])
                    tbl_lst[i + j].insert(y[1], y[2])

    # if row_spanned:
    #     for x in row_spanned:
    #         for j in range(1, x[2]):
    #             # Add value in next tr
    #             idx = x[0] + j
    #             # assert isinstance(idx, int)
    #             if x[1] >= len(tbl_lst[idx]):
    #                 tbl_lst[idx].insert(x[1], x[3])
    #             elif x[3] in tbl_lst[x[0]]:
    #                 tbl_lst[idx].insert(tbl_lst[x[0]].index(x[3]), x[3])
    #             else:
    #                 tbl_lst[idx].insert(x[1] + 1, x[3])

    for k in range(len(tbl_lst)):
        n = len(header) - len(tbl_lst[k])
        if n > 0:
            tbl_lst[k].extend(['\xa0'] * n)
        elif n < 0 and tbl_lst[k][2] == '\xa0':
            del tbl_lst[k][2]

    return tbl_lst


# Parse the acquired list to make it be ready for creating the DataFrame
def parse_table(source, parser='lxml'):
    """
    :param source: response object to connecting a URL to request a table
    :param parser: [str] Optional parsers: 'lxml', 'html5lib', 'html.parser'
    :return [tuple] ([list] of lists each comprising a row of the requested table - (see also parse_trs())
                     [list] of column names of the requested table)
    """
    # (If source.status_code == 200, the requested URL is available.)
    # Get plain text from the source URL
    web_page_text = source.text
    # Parse the text
    parsed_text = bs4.BeautifulSoup(web_page_text, parser)
    # Get all data under the HTML label 'tr'
    table_temp = parsed_text.find_all('tr')
    # Get a list of column names for output DataFrame
    headers = table_temp[0]
    header = [header.text for header in headers.find_all('th')]
    # Get a list of lists, each of which corresponds to a piece of record
    trs = table_temp[1:]
    # Return a list of parsed tr's, each of which corresponds to one df row
    return parse_tr(header, trs), header


# Parse location note
def parse_location_note(x_note):
    # Data
    # d = re.search('[\w ,]+(?=[ \n]\[)', x)
    # if d is not None:
    #     dat = d.group()
    # else:

    # Location name
    d = re.search(r'.*(?= \[[\"\']\()', x_note)
    if d is not None:
        dat = d.group()
    elif ' [unknown feature, labelled "do not use"]' in x_note:
        dat = re.search(r'\w+(?= \[unknown feature, )', x_note).group()
    elif ') [formerly' in x_note:
        dat = re.search(r'.*(?= \[formerly)', x_note).group()
    else:
        m_pattern = re.compile(
            r'[Oo]riginally |[Ff]ormerly |[Ll]ater |[Pp]resumed | \(was | \(in | \(at | \(also |'
            r' \(second code |\?|\n| \(\[\'| \(definition unknown\)')
        # dat = re.search('["\w ,]+(?= [[(?\'])|["\w ,]+', x).group(0) if re.search(m_pattern, x) else x
        x_tmp = re.search(r'(?=[\[(]).*(?<=[\])])|(?=\().*(?<=\) \[)', x_note)
        x_tmp = x_tmp.group() if x_tmp is not None else x_note
        dat = ' '.join(x_note.replace(x_tmp, '').split()) if re.search(m_pattern, x_note) else x_note

    # Note
    y = x_note.replace(dat, '').strip()
    if y == '':
        note = ''
    else:
        n = re.search(r'(?<=[\[(])[\w ,?]+(?=[])])', y)
        # n = re.search('(?<=[\n ]((\[\'\()|(\(\[\')))[\w ,\'\"/?]+', y)
        if n is None:
            n = re.search(r'(?<=(\[[\'\"]\()|(\([\'\"]\[)|(\) \[)).*(?=(\)[\'\"]\])|(\][\'\"]\))|\])', y)
        elif '"now deleted"' in y and y.startswith('(') and y.endswith(')'):
            n = re.search(r'(?<=\().*(?=\))', y)
        note = n.group() if n is not None else ''
        if note.endswith('\'') or note.endswith('"'):
            note = note[:-1]

    if 'STANOX ' in dat and 'STANOX ' in x_note and note == '':
        dat = x_note[0:x_note.find('STANOX')].strip()
        note = x_note[x_note.find('STANOX'):]

    return dat, note


# Parse date string
def parse_date(str_date, as_date_type=False):
    """
    :param str_date: [str]
    :param as_date_type: [bool]
    :return: the date formatted as requested
    """
    parsed_date = dateutil.parser.parse(str_date, fuzzy=True)
    # Or, parsed_date = datetime.strptime(last_update_date[12:], '%d %B %Y')
    parsed_date = parsed_date.date() if as_date_type else str(parsed_date.date())
    return parsed_date


# ====================================================================================================================
""" Get useful information """


# Get last update date
def get_last_updated_date(url, parsed=True, date_type=False):
    """
    :param url: [str] URL link of a requested web page
    :param parsed: [bool] indicator of whether to reformat the date
    :param date_type: [bool]
    :return:[str] date of when the specified web page was last updated
    """
    # Request to get connected to the given url
    source = requests.get(url)
    web_page_text = source.text
    # Parse the text scraped from the requested web page
    # (Optional parsers: 'lxml', 'html5lib' and 'html.parser')
    parsed_text = bs4.BeautifulSoup(web_page_text, 'lxml')
    # Find 'Last update date'
    update_tag = parsed_text.find('p', {'class': 'update'})
    if update_tag is not None:
        last_update_date = update_tag.text
        # Decide whether to convert the date's format
        if parsed:
            # Convert the date to "yyyy-mm-dd" format
            last_update_date = parse_date(last_update_date, date_type)
    else:
        last_update_date = None  # print('Information not available.')
    return last_update_date


#
def get_navigation_elements(lst):
    """
    :param lst: [list]
    :return:
    """
    assert isinstance(lst, list)
    lst_reversed = list(reversed(lst))
    for x in lst_reversed:
        if x.text == 'Introduction':
            starting_idx = len(lst) - lst_reversed.index(x) - 1
            return lst[starting_idx:]


#
def get_catalogue(url, navigation_bar_exists=True, menu_exists=True):
    """
    :param url: [str]
    :param navigation_bar_exists: [bool]
    :param menu_exists: [bool]
    :return:
    """

    source = requests.get(url)

    if navigation_bar_exists:
        cold_soup = bs4.BeautifulSoup(source.text, 'lxml').find_all('a', text=True, attrs={'class': None})
        hot_soup = get_navigation_elements(cold_soup)
    else:
        if menu_exists:
            hot_soup = bs4.BeautifulSoup(source.text, 'lxml').find_all('a', text=True, attrs={'class': None})[-6:]
        else:
            hot_soup = []

    source.close()

    raw_contents = [{x.text: urllib.parse.urljoin(os.path.dirname(url) + '/', x['href'])} for x in hot_soup]

    contents = dict(e for d in raw_contents for e in d.items())

    return contents


#
def get_cls_menu(cls_url):
    source = requests.get(cls_url)

    soup = bs4.BeautifulSoup(source.text, 'lxml')
    h1, h2s = soup.find('h1'), soup.find_all('h2')

    cls_name = h1.text.replace(' menu', '')

    if len(h2s) == 0:
        cls_elem = dict((x.text, urllib.parse.urljoin(cls_url, x.get('href'))) for x in h1.find_all_next('a'))

    else:
        all_next = [x.replace(':', '') for x in h1.find_all_next(string=True) if x != '\n' and x != '\xa0'][2:]
        h2s_list = [x.text.replace(':', '') for x in h2s]
        all_next_a = [(x.text, urllib.parse.urljoin(cls_url, x.get('href'))) for x in h1.find_all_next('a', href=True)]

        idx = [all_next.index(x) for x in h2s_list]
        for i in idx:
            all_next_a.insert(i, all_next[i])

        cls_elem, i = {}, 0
        while i <= len(idx):
            if i == 0:
                d = dict(all_next_a[i:idx[i]])
            elif i < len(idx):
                d = {h2s_list[i-1]: dict(all_next_a[idx[i-1]+1:idx[i]])}
            else:
                d = {h2s_list[i-1]: dict(all_next_a[idx[i-1]+1:])}
            i += 1
            cls_elem.update(d)

    cls_menu = {cls_name: cls_elem}
    return cls_menu


# ====================================================================================================================
""" Misc """


#
def is_float(text):
    try:
        float(text)  # float(re.sub('[()~]', '', text))
        test_res = True
    except ValueError:
        test_res = False
    return test_res



""" Railway tunnel lengths """

import copy
import itertools
import operator
import os
import re

import bs4
import fuzzywuzzy.process
import measurement.measures
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_tr
from pyrcs.utils import save_pickle


class Tunnels:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Tunnels'
        self.URL = self.HomeURL + '/tunnels/tunnels0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other_assets", 'tunnels')
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other_assets\\tunnels" and sub-directories
    def cd_tunnels(self, *directories):
        path = self.DataDir
        for x in directories:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other_assets\\tunnels\\dat" and sub-directories
    def cdd_tunnels(self, *sub_dir):
        path = self.cd_tunnels("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Find page headers
    def find_page_headers(self):
        intro_page = requests.get(self.URL)
        pages = [x.text for x in bs4.BeautifulSoup(intro_page.text, 'lxml').find_all('a', text=re.compile('^Page.*'))]
        return pages[:int(len(pages) / 2)]

    # Parse data in 'Length' column, i.e. convert miles/yards to metres
    @staticmethod
    def parse_tunnel_length(x):
        """
        :param x:
        :return:

        Examples:
            '' or Unknown
            1m 182y; 0m111y; c0m 150y; 0m 253y without avalanche shelters
            0m 56ch
            formerly 0m236y
            0.325km (0m 356y); 0.060km ['(0m 66y)']
            0m 48yd- (['0m 58yd'])
        """
        if re.match(r'[Uu]nknown', x):
            length = pd.np.nan
            add_info = 'Unknown'
        elif x == '':
            length = pd.np.nan
            add_info = 'Unavailable'
        elif re.match(r'\d+m \d+yd-.*\d+m \d+yd.*', x):
            miles_a, yards_a, miles_b, yards_b = re.findall(r'\d+', x)
            length_a = measurement.measures.Distance(mi=miles_a).m + measurement.measures.Distance(yd=yards_a).m
            length_b = measurement.measures.Distance(mi=miles_b).m + measurement.measures.Distance(yd=yards_b).m
            length = (length_a + length_b) / 2
            add_info = '-'.join([str(round(length_a, 2)), str(round(length_b, 2))]) + ' metres'
        else:
            if re.match(r'(formerly )?c?\d+m ?\d+y?(ch)?.*', x):
                miles, yards = re.findall(r'\d+', x)
                if re.match(r'.*\d+ch$', x):
                    yards = measurement.measures.Distance(chain=yards).yd
                if re.match(r'^c.*', x):
                    add_info = 'Approximate'
                elif re.match(r'\d+y$', x):
                    add_info = re.search(r'(?<=\dy).*$', x).group()
                elif re.match(r'^(formerly).*', x):
                    add_info = 'Formerly'
                else:
                    add_info = None
            elif re.match(r'\d+\.\d+km .*\(\d+m \d+y\).*', x):
                miles, yards = re.findall(r'\d+', re.search(r'(?<=\()\d+.*(?=\))', x).group())
                add_info = re.search(r'.+(?= (\[\')?\()', x).group()
            else:
                print(x)
                miles, yards = 0, 0
                add_info = ''
            length = measurement.measures.Distance(mi=miles).m + measurement.measures.Distance(yd=yards).m
        return length, add_info

    # Collect the data of railway tunnel lengths for a given page number
    def collect_railway_tunnel_lengths(self, page_no, update=False, verbose=False):
        """
        :param page_no: [int] page number; valid values include 1, 2, and 3
        :param update: [bool] indicate whether to re-scrape the tunnel lengths data for the given page_no
        :param verbose: [bool]
        :return [dict] containing:
                    [DataFrame] tunnel lengths data of the given 'page'
                    [str] date of when the data was last updated
        """
        assert page_no in range(1, 4)
        page_headers = self.find_page_headers()
        filename = fuzzywuzzy.process.extractOne(str(page_no), page_headers)[0]
        pickle_filename = re.sub(r"[()]", "", re.sub(r"[ -]", "_", filename)).lower() + ".pickle"

        path_to_pickle = self.cd_tunnels(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            tunnel_lengths_codes = load_pickle(path_to_pickle)

        else:
            url = self.URL.replace('tunnels0', 'tunnels{}'.format(page_no))

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last updated date for tunnel lengths data on \"{}\". {}".format(filename, e))
                last_updated_date = ''

            try:
                source = requests.get(url)
                parsed_text = bs4.BeautifulSoup(source.text, 'lxml')  # Optional parsers:, 'html5lib', 'html.parser'

                header = [x.text for x in parsed_text.find_all('th')]  # Column names

                crossed = [re.match(r'^Between.*', x) for x in header]
                if any(crossed):
                    idx = list(itertools.compress(range(len(crossed)), crossed))
                    assert len(idx) == 1
                    header.remove(header[idx[0]])
                    header[idx[0]:idx[0]] = ['Station_O', 'Station_D']

                # Table data
                temp_tables = parsed_text.find_all('table', attrs={'width': '1100px'})
                tbl_lst = parse_tr(header, trs=temp_tables[1].find_all('tr'))
                tbl_lst = [[item.replace('\r', ' ').replace('\xa0', '') for item in record] for record in tbl_lst]

                # Create a DataFrame
                tunnel_lengths_table = pd.DataFrame(data=tbl_lst, columns=header)
                tunnel_lengths_table[['Length_m', 'Length_note']] = \
                    tunnel_lengths_table.Length.map(self.parse_tunnel_length).apply(pd.Series)

            except Exception as e:
                print("Failed to collect tunnel lengths data on \"{}\". {}".format(filename, e))
                tunnel_lengths_table = pd.DataFrame()

            code_key = re.search(r'(?<=\()\w.*(?=\))', filename).group(0).replace('-', '_')
            tunnel_lengths_codes = {code_key: tunnel_lengths_table, 'Last_updated_date': last_updated_date}

            save_pickle(tunnel_lengths_codes, path_to_pickle, verbose)

        return tunnel_lengths_codes

    # Collect the data of minor lines and other odds / ends
    def collect_page4_others(self, update=False, verbose=False):
        """
        Page 4 (others) contains more than one table on the web page
        """
        page_headers = self.find_page_headers()
        filename = fuzzywuzzy.process.extractOne('others', page_headers)[0]
        pickle_filename = re.sub(r"[()]", "", re.sub(r"[ -]", "_", filename)).lower() + ".pickle"
        path_to_pickle = self.cd_tunnels(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            tunnel_lengths_data = load_pickle(path_to_pickle)

        else:
            url = self.HomeURL + '/tunnels/tunnels4.shtm'

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last updated date for tunnel lengths data on \"{}\". {}".format(filename, e))
                last_updated_date = ''

            try:
                source = requests.get(url)
                parsed_text = bs4.BeautifulSoup(source.text, 'lxml')  # Optional parsers:, 'html5lib', 'html.parser'
                headers = []
                temp_header = parsed_text.find('table')
                while temp_header.find_next('th'):
                    header = [x.text for x in temp_header.find_all('th')]
                    if len(header) > 0:
                        crossed = [re.match('^Between.*', x) for x in header]
                        if any(crossed):
                            idx = list(itertools.compress(range(len(crossed)), crossed))
                            assert len(idx) == 1
                            header.remove(header[idx[0]])
                            header[idx[0]:idx[0]] = ['Station_O', 'Station_D']
                        headers.append(header)
                    temp_header = temp_header.find_next('table')

                tbl_lst = parsed_text.find_all('table', attrs={'width': '1100px'})
                tbl_lst = operator.itemgetter(1, 3)(tbl_lst)
                tbl_lst = [parse_tr(header, x.find_all('tr')) for header, x in zip(headers, tbl_lst)]
                tbl_lst = [[[item.replace('\xa0', '') for item in record] for record in tbl] for tbl in tbl_lst]

                tunnel_lengths = [pd.DataFrame(tbl, columns=header) for tbl, header in zip(tbl_lst, headers)]
                for i in range(len(tunnel_lengths)):
                    tunnel_lengths[i][['Length_m', 'Length_note']] = \
                        tunnel_lengths[i].Length.map(self.parse_tunnel_length).apply(pd.Series)

                other_types = [x.text for x in parsed_text.find_all('h3')]

            except Exception as e:
                print("Failed to collect tunnel lengths data on \"{}\". {}".format(filename, e))
                other_types, tunnel_lengths = ['None'], pd.DataFrame()

            tunnel_lengths_data = dict(zip(other_types + ['Last_updated_date'], tunnel_lengths + [last_updated_date]))

            save_pickle(tunnel_lengths_data, path_to_pickle, verbose)

        return tunnel_lengths_data

    # Fetch all the collected data of railway tunnel lengths
    def fetch_railway_tunnel_lengths(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param update: [bool]
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param verbose: [bool]
        :return [dict] containing:
                    [DataFrame] railway tunnel lengths data, including the name,
                                length, owner and relative location
                    [str] date of when the data was last updated
        """
        data_sets = [self.collect_railway_tunnel_lengths(
            page_no, update, verbose=False if data_dir or not verbose else True) for page_no in range(1, 4)]
        other_data_set = self.collect_page4_others(update, verbose=False if data_dir or not verbose else True)

        tunnel_lengths_tables = [d[k] for d in data_sets for k, v in d.items() if k != 'Last_updated_date']
        others_tables = [v for k, v in other_data_set.items() if k != 'Last_updated_date']
        tunnel_lengths_tables += others_tables
        tunnel_lengths_codes = pd.concat(tunnel_lengths_tables, axis=0, ignore_index=True, sort=False)

        last_updated_dates = [d[k] for d in data_sets for k, v in d.items() if k == 'Last_updated_date']
        latest_update_date = max(last_updated_dates)

        tunnel_lengths = {'Tunnels': tunnel_lengths_codes, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "tunnel_lengths.pickle")
            save_pickle(tunnel_lengths, path_to_pickle, verbose=True)

        return tunnel_lengths



""" Settings """


# Set preferences for displaying results
def np_preferences(reset=False):
    import numpy as np
    if not reset:
        np.core.arrayprint._line_width = 120
    else:
        np.core.arrayprint._line_width = 80  # 75


# Set preferences for displaying results
def pd_preferences(reset=False):
    import pandas as pd
    if not reset:
        # pandas.set_option('display.float_format', lambda x: '%.4f' % x)
        pd.set_option('display.precision', 2)
        pd.set_option('expand_frame_repr', False)  # Set the representation of DataFrame NOT to wrap
        pd.set_option('display.width', 600)  # Set the display width
        pd.set_option('precision', 4)
        pd.set_option('display.max_columns', 100)
        pd.set_option('display.max_rows', 20)
        pd.set_option('io.excel.xlsx.writer', 'xlsxwriter')
        pd.set_option('mode.chained_assignment', None)
        # pandas.set_option('display.float_format', lambda x: '%.4f' % x)
    else:
        pd.reset_option('all')


# Set preferences for plotting
def mpl_preferences(use_cambria=False, reset=False):
    """
    Get a list of supported file formats for matplotlib savefig() function
      plt.gcf().canvas.get_supported_filetypes()  # Aside: "gcf" is short for "get current fig" manager
      plt.gcf().canvas.get_supported_filetypes_grouped()
    """
    import matplotlib.pyplot as plt
    if not reset:
        if use_cambria:  # Use the font, 'Cambria'
            # Add 'Cambria' and 'Cambria Math' to the front of the 'font.serif' list
            plt.rcParams['font.serif'] = ['Cambria'] + plt.rcParams['font.serif']
            plt.rcParams['font.serif'] = ['Cambria Math'] + plt.rcParams['font.serif']
            # Set 'font.family' to 'serif', so that matplotlib will use that list
            plt.rcParams['font.family'] = 'serif'
        plt.rcParams['font.size'] = 13
        plt.rcParams['font.weight'] = 'normal'
        plt.rcParams['legend.labelspacing'] = 0.7
        plt.style.use('ggplot')
    else:
        plt.rcParams = plt.rcParamsDefault
        plt.style.use('classic')


# Set GDAL configurations
def gdal_configurations(reset=False, max_tmpfile_size=2500):
    import gdal
    if not reset:
        # Whether to enable interleaved reading. Defaults to NO.
        gdal.SetConfigOption('OGR_INTERLEAVED_READING', 'YES')
        # Whether to enable custom indexing. Defaults to YES.
        gdal.SetConfigOption('USE_CUSTOM_INDEXING', 'YES')
        # Whether to compress nodes in temporary DB. Defaults to NO.
        gdal.SetConfigOption('COMPRESS_NODES', 'YES')
        # Maximum size in MB of in-memory temporary file. If it exceeds that value, it will go to disk. Defaults to 100.
        gdal.SetConfigOption('MAX_TMPFILE_SIZE', str(max_tmpfile_size))
    else:
        gdal.SetConfigOption('OGR_INTERLEAVED_READING', 'NO')
        gdal.SetConfigOption('USE_CUSTOM_INDEXING', 'YES')
        gdal.SetConfigOption('COMPRESS_NODES', 'NO')
        gdal.SetConfigOption('MAX_TMPFILE_SIZE', '100')



import os

import setuptools

ver = '0.2.0'


def find_all_pkg_dat_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


dat_files = find_all_pkg_dat_files('pyrcs\\dat')


with open("README.md", 'r', encoding='utf-8') as readme:
    long_description = readme.read()

setuptools.setup(

    name='pyrcs',
    version=ver,

    author='Qian Fu',
    author_email='qian.fu@outlook.com',

    description="A small web scraper for collecting the railway codes used in the UK rail industry",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url='https://github.com/mikeqfu/pyrcs',

    install_requires=[
        'beautifulsoup4',
        'fuzzywuzzy',
        'html5lib',
        'lxml',
        'measurement',
        'more-itertools',
        'pandas',
        'pyhelpers',
        'python-Levenshtein',
        'requests',
        'sqlalchemy',
        'sqlalchemy-utils'
    ],

    packages=setuptools.find_packages(exclude=["*.tests", "tests.*", "tests"]),

    package_data={"pyrcs": dat_files + ["line_data_cls/*", "other_assets_cls/*"]},
    include_package_data=True,

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: Microsoft :: Windows',
    ],
)
